package heap;

// The heapsort algorithm - class HeapArray.

// This implementation is modeled after the courseware Heapsort-SALA.
// http://OLLI.Informatik.Uni-Oldenburg.DE/heapsort_SALA/
// It is optimized for easy understanding and not for efficiency.
// (c) 2001 Nils Faltin. Distributed under the GNU General Public License. 
// The license text ist contained in the courseware distribution.

// used by class HeapsortArray
// contains heap tree and result list
// provides 
// - specialized access methods to nodes
// - test for emptiness of heaptree
// - swap 

abstract public class HeapArray extends NodeArray{
   final static int rootIndex = 0; // Heap tree root stored at index 0.
   final static int nullIndex = rootIndex - 1;
   int lastHeap = nullIndex; // Points to last Node in Heap. Initialize for empty heap.
   

	// class methods
   // constructor: initialize with empty array of length len
	public HeapArray(int len){
	  super(len);
	}
	
	// instance methods	
    // add a Node
    public void addNode(Node n){
      super.addNode(n);
      incrementHeapsize(); 
    }

   // for debug: return heapTree formatted as a tree
   protected StringBuffer heapTreeString(){
      int ns = rootIndex; // index of start node in current level
      int ne = leftIndex(ns); // index of start node in next level
      int spa = 7; // number of spaces before first node
      int spb = 1; // number of spaces between nodes
      
      StringBuffer s = new StringBuffer();
      for (int level = 0; level <= 3; level ++){
         s.append(space(spa)); // space before first node
         for (int ni=ns; ni < ne; ni++){ // ni: inices of nodes in current level
           String node = at(ni) != nullIndex? getNodeAt(ni).toString() : " "; // get Nodes string representation; space if no node any more
           s.append(node + space(spb)); // append node and space between nodes
         } 
         s.append("\n"); // new line
         ns = leftIndex(ns) ;
         ne = leftIndex(ne) ;
         spb = spa;
         spa = (spa+1) / 2 -1;
      }  
      return s;
   }
   
   // for debug: return a string of len spaces
   protected String space(int len){
     return "                                                                                   ".substring(1,len+1);
   }
   
   // for debug: left index of child node
   private int leftIndex(int k){
     return 2 * k + 1; // Because root is at 0. Root at 1 gives 2*k.
   }
                  
   // algorithm functions
   
    // heaporder the whole tree
    protected void buildHeap(){ 
	  // Start at last non-leaf node. Process all nodes in reverse storage order untill we reach the root
		// by this order, we ensure that heapify's precondition is fulfilled
	  for (int k = lastInnerNode(); k != nullIndex; k=predesessor(k)){
	    heapify(k);
      }
    }

	// Heapify the subtree located at k
	// Precondition: k's both childtrees are heapordered
    protected void heapify(int k){ 
		// move the key down the tree till we're done.
		while( k != nullIndex){
		  k = heapifyLocally(k);
		};     
    }
    
	// heap order Node k with respect to its both children
	// if keys had to be swapped return the node where k.key now is, nullIndex otherwise
    protected int heapifyLocally(int k){
       int mc = maxChild(k);
       if (mc == nullIndex){ return nullIndex;} // k is leaf. we're done.
       if (isGreaterThan(mc,k)){ // if max child has bigger key then swap
         swap(k,mc);
         return mc;
       }else{
         return nullIndex;
       }
    }

	// move the maximum node from the heap into the result list
	// precondition: the tree must be a heap
	// return index where node was placed in result list
   protected int moveMax(){
      swap(rootIndex,lastHeap); // move maximum node to result list
      decrementHeapsize(); // heap is now one Node smaller
      heapify(rootIndex); // restore the heap
      return lastHeap + 1;
   }
   

   // heap tree methods
   
   
   // is the heap empty?
   protected boolean heapEmpty(){
     return lastHeap == nullIndex;
   }
   
   // A node has been moved out of the heap. Register this by decrementing the heapsize counter.
   protected void decrementHeapsize(){
	 lastHeap--;
   }

   protected void incrementHeapsize(){
	 lastHeap++;
   }

   // node access methods

   // return last inner Node (has no children) in Heap , nullIndex if there is no such node 
   protected int lastInnerNode(){
     return heapEmpty()? nullIndex : parent(lastHeap);
   }
   
   // if index is in heap, return it unchanged, otherwise return nullIndex
   protected int at(int index){
     return index >=rootIndex && index <= lastHeap ? index : nullIndex;
   }
   
   // return the predessor of Node k in storage order, nullIndex if k is the root
   protected int predesessor(int k){
     return at(k - 1);
   }
   
   // Return left child of Node k, nullIndex if there is no such child
   protected int leftChild(int k){
     return at(2 * k + 1); // Because root is at 0. Root at 1 gives 2*k.
   }
   
   // Return rihgt child of Node k, nullIndex if there is no such child
   protected int rightChild(int k){
     return at(2 * k + 2); // Because root is at 0. Root at 1 gives 2*k +1.
   }
   
   // return parent node of node k. return nullIndex if k is the root.
   protected int parent(int k){
     return k== rootIndex? nullIndex : at((k-1) / 2); // Because root is at 0. Root at 1 gives k/2.
     	// parent(root) is now root. This makes ?: necessary
   }
   
   // return the bigger child of Node k. Return nullIndex if k is a leaf.
   // precondition: heap must be complete
   protected int maxChild(int k){
     int rc = rightChild(k);
     int lc = leftChild(k);
     
     if (rc == nullIndex){
       return lc;
     }else{ // because heap is complete there must be a left child
        return isGreaterThan(lc, rc)? lc : rc;
     }
   }
}
   
